/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2000 France Telecom R&D
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Fabien Delpiano, Bruno Dumant
 * 
 * -----------------------------------------------------------------
 * $Id: KKernel.java,v 1.1.1.1 2002/08/06 13:14:03 bmi Exp $
 * -----------------------------------------------------------------
 */

package org.objectweb.jonathan.libs.kernel;

import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.ConfigurationFactory;
import org.objectweb.jonathan.apis.kernel.JonathanException;
import org.objectweb.jonathan.apis.kernel.InternalException;
import org.objectweb.jonathan.apis.kernel.ClassPathFinder;
import org.objectweb.jonathan.apis.kernel.decoders.Properties2Configuration;
import org.objectweb.jonathan.apis.kernel.decoders.File2Configuration;

import java.io.IOException;
import java.util.Properties;

/**
 * Bootstrap class of the jonathan configuration mechanism.
 * Provides the runtime configuration file mechanism by using
 * XML decoders and loads the java system properties.<P>
 * Must be provided in the application's compiled configuration
 * by specifying the <B>bootstrap</B> attribute of the root
 * context in the XML configuration file to be compiled. KKernel
 * is loaded in the java code of the compiled configuration file.
 * <P>
 * <U>jonathan's example of KKernel use:</U>
 * <ul>
 *   <li>the compiled configuration file is Kernel.java; it is
 * obtained from its XML source Kernel.kcf by the {@link org.objectweb.jonathan.tools.Kcf2java}
 * tool. Its class name (Kernel) is provided in the XML source
 * by the <b>generated</b> attribute of the root context.
 *   <li>the runtime configuration file is jonathan.xml; its is
 * provided as the property named
 * <I>"/jonathan/KKernel/runtime_configuration_file"</I> in the
 * Kernel.kcf.
 *   <li>the runtime configuration file is decoded by an
 * implementation of the interface {@link File2Configuration} (in fact
 * an {@link org.objectweb.jonathan.libs.kernel.decoders.OpenXMLDecoder}) whose configuration is provided in
 * Kernel.kcf.
 * as an assemblage named <I>/jonathan/KKernel/xml_decoder</I>.
 * The decoder is instanciated by the configuration bootstrap class
 * {@link KKernel}.
 *   <li>the whole mechanism is triggered by the {@link org.objectweb.david.libs.binding.orbs.ORBClass}
 * <code>setParameters()</code> method, which calls the static
 * Kernel.newConfiguration() method and returns jonathan's
 * whole runtime configuration.
 * </ul>
 */
public final class KKernel {

   static Container containers = null;
   static ConfigurationFactory configuration_factory = new JConfigurationFactory();
   
   /** 
    * The jonathan Kernel's runtime bootstrap mechanism.
    * This method
    * <UL>
    *   <li>loads its context parameter's
    * <I>/jonathan/KKernel/xml_decoder</I> value; it must be an
    * implementation of {@link File2Configuration}
    *   <li>loads its context parameter's
    * <i>/jonathan/KKernel/runtime_configuration_file</i> String
    * property and decodes the corresponding runtime configuration
    * file with the previously loaded decoder; the decoded components
    * are imported in a duplicata of the given Context parameter.
    *   <li>loads its context parameter's
    * <I>/jonathan/KKernel/properties_decoder</I> (it must be an
    * implementation of {@link Properties2Configuration}) and uses
    * it to import the java System properties
    * </UL>
    *
    * @param _context the compiled configuration to bootstrap on
    * @param _c
    * @return the application's complete runtime configuration
    */
   public static Context newConfiguration(Context _context, Class _c){
      ClassLoader $loader = null;
      try {
         $loader = _c.getClassLoader();
      } catch (Exception e) {
         e.printStackTrace();
      }

      Container $current = containers;
      while ($current != null) {
         if ($loader == $current.loader && _context == $current.source) {
            return $current.returned_context;
         } else {
            $current = $current.next;
         }
      }

      ClassPathFinder $class_path_finder = new ClassPathFinderImpl($loader);

      Context $toBeReturned =
         configuration_factory.newContext(_context,$class_path_finder);

      Object $xml_file_name = null;
      try {
         $xml_file_name = System.getProperty("jonathan.KKernel.runtime_configuration_file");
      } catch (Exception _) {
      }
      
      if ($xml_file_name == null) {
          $xml_file_name =
             _context.getValue("/jonathan/KKernel/runtime_configuration_file",'/');
      }

      if ($xml_file_name instanceof String) {
         Object $component = $toBeReturned.getValue("/jonathan/KKernel/xml_decoder",'/');
         if ($component instanceof File2Configuration) {
            File2Configuration $xml_decoder = (File2Configuration) $component;
            try {
               $xml_decoder.importElement($toBeReturned,(String) $xml_file_name,configuration_factory,
                                          $class_path_finder);
            } catch (JonathanException e) {
               Exception actual = e.represents();
               if (actual instanceof IOException) {
                  System.err.println("Warning : error when trying to import the runtime configuration from " + 
                                     $xml_file_name +": " + e);
                  $toBeReturned = configuration_factory.newContext(_context,$class_path_finder);
               } else {
                  System.err.println("Error importing the runtime configuration from " + 
                                     $xml_file_name +".");
                  e.printStackTrace();
               }
            }
         } else {
            System.err.println("Warning: no XML decoder associated with KKernel to read input from " +
                               $xml_file_name);
         }
      }
      
      Properties system_properties = null;
      try {
         system_properties = System.getProperties();
      } catch (SecurityException ignored) {
         ignored.printStackTrace();
      }

      if (system_properties != null) {
         Object $component = $toBeReturned.getValue("/jonathan/KKernel/properties_decoder",'/');
         if ($component instanceof Properties2Configuration) {
            Properties2Configuration $properties_decoder = (Properties2Configuration) $component;
            try {
               $properties_decoder.importProperties($toBeReturned, system_properties, 
                                                    new String[] {"java.", "file."}, true);
            } catch (JonathanException e) {
               System.err.println("Warning : error when trying to import system properties: " + e);
            }
         } else {
            System.err.println("Warning: no properties decoder associated with KKernel to\n" + 
                               "         import the system properties.");
         }
      }
      
      // -------------------

      containers = new Container(_context,$loader,$toBeReturned,containers);
      return $toBeReturned;
   }

   static class Container {
      Context source;
      ClassLoader loader;
      Context returned_context;
      Container next;

      Container(Context _source,ClassLoader _loader,Context _returned_context,
                Container _next) {
         source = _source;
         loader = _loader;
         returned_context = _returned_context;
         next = _next;
      }
   }
}   



