/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2000 France Telecom R&D
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Fabien Delpiano, Bruno Dumant
 * 
 * -----------------------------------------------------------------
 * $Id: TreeContextFactory.java,v 1.1.1.1 2002/08/06 13:13:50 bmi Exp $
 * -----------------------------------------------------------------
 */

package org.objectweb.jonathan.libs.kernel;

import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.ContextFactory;
import org.objectweb.jonathan.apis.kernel.Element;
import org.objectweb.jonathan.apis.kernel.Component;
import org.objectweb.jonathan.apis.kernel.Name;
import org.objectweb.jonathan.apis.kernel.JonathanException;
import org.objectweb.jonathan.apis.kernel.InternalException;

import org.objectweb.jonathan.libs.helpers.NameHelpers;

public class TreeContextFactory extends JContextFactory {

   static final Name SELF_NAME = NameHelpers.newName(".",(char) 0);
   static final Name ROOT_NAME = NameHelpers.newName("",(char) 0);

   static TreeContext[] pool = new TreeContext[10];
   static int pool_size = 0;

   /** 
    * Indicates whether type checking should be performed.
    */
   public static final boolean check_types = false;
   public static final TreeContextFactory instance = new TreeContextFactory();

   /** 
    * Returns a new {@link TreeContextFactory.TreeContext TreeContext} 
    * instance.
    * @return a new {@link TreeContextFactory.TreeContext} instance.
    */
   public Context newContext() {
      synchronized (TreeContextFactory.class) {
         if (pool_size > 0) {
            TreeContext context = pool[--pool_size];
            context.acquire();
            return context;
         } else {
            return new TreeContext();
         }
      }
   }

   /** 
    * Returns a new {@link TreeContextFactory.TreeContext TreeContext} 
    * instance with scope 
    * <code>scope</code>. 
    * <p>
    * <code>_scope</code> must have been acquired by the caller first, if it is
    * to be shared.
    *
    * @param scope the scope of the returned context instance;
    * @return a new {@link TreeContextFactory.TreeContext TreeContext} 
    *         instance.
    */
   public Context newContext(Context _scope) {
      synchronized (TreeContextFactory.class) {
         if (pool_size > 0) {
            TreeContext context =  pool[--pool_size];
            context.acquire();
            context.enclosing_context = _scope;
            return context;
         } else {
            return new TreeContext(_scope);
         }
      }
   }

   protected class TreeContext extends JContext {

      /** element representing the current context. Initialized by the constructors. */
      protected Element self_element;

      /** 
       * Constructs a new TreeContext instance.
       */
      protected TreeContext() {
         super();
         self_element = new JElement(this,".",this);
      }
      
      protected TreeContext(Context _context) {
         super(_context);
         self_element = new JElement(this,".",this);
      }
      
      public String toString() {
         return "TreeContext[" + id + "," + getAbsoluteName() + 
            (enclosing_context != null ? "," + enclosing_context + "]]" : "]");
      }
      
      public Name getAbsoluteName() {
         if (reference instanceof TreeElement) {
            return ((TreeElement) reference).getAbsoluteName();
         } else if (reference == null) {
            return NameHelpers.newName("",null);
         } else {
            return NameHelpers.newName("<undefined>",null);
         }
      }
      

      protected void recycle() {
         synchronized (TreeContextFactory.class) {
            if (pool_size < pool.length) {
               pool[pool_size++] = this;
            }
         }
      }

      /** 
       * Returns the element identified by name, or null if none exists.
       *
       * @param name
       * @param local if true, don't search the enclosing context.
       * @return the element identified by name, or null if none exists.
       */
      protected Element getElement(String _name,boolean _local) {
         if (".".equals(_name)) {
            return self_element; 
         } else if ("..".equals(_name)) {
            return reference != null ? reference.getContainer().getElement(SELF_NAME) : null;
         } else if ("".equals(_name)) {
            if (reference != null) {
               try {
                  return reference.getContainer().getElement(ROOT_NAME);
               } catch (NullPointerException e) {
                  System.err.println(reference);
                  throw e;
               }
            } else {
               return self_element;
            }
         } else {
            return super.getElement(_name,_local);
         }
      }

      /** 
       * Adds an Element to the Context if no other exists with the same name.
       * <p>
       * must be  called in a synchronized context
       * <p>
       * @param _name the name of the new Element.
       * @param _type the declared type of the new Element.
       * @param _value the value of the Element to be inserted.
       * @param _int_value the integer value of the Element to be inserted.
       * @param _check_types indicates whether types should be checked.
       * @return the Element already present in the Context with the same name, 
       *         null if none existed.
       * @exception JonathanException if something wrong occurs.
       */
      protected Element addLocalElement(ContextElement _element,
                                        boolean _check_types)
         throws JonathanException {
         //System.err.println("Adding " + _element + " to " + this);
         String $name = _element.name;
         Component $component = _element.component; 
         if ($name.equals("..")) {
            if ($component != null) {
               throw new JonathanException("Can't add the parent context.");
            }
         } else if ($name.equals(".")) {
            throw new JonathanException("Can't add an element named \".\".");
         }
         // actual addition
         return super.addLocalElement(_element,_check_types);
      }

      public Component duplicate() {
         return TreeContextFactory.this.newContext(this);
      }

      protected ContextElement newContextElement(Context _container) {
         ContextElement result = null;
         if (unused != null) {
            result = unused;
            unused = unused.next;
            result.next = null;
            result.container = _container;
         } else {
            result = new TreeContextElement(_container);
         }
         return result;
      }
   }

   static class TreeContextElement extends ContextElement implements TreeElement {
      
      public TreeContextElement(Context container) {
         super(container);
      }
      
      public Name getAbsoluteName() {
         return getAbsoluteName(null);
      }
      
      public Name getAbsoluteName(Name subname) {
         Name new_name = NameHelpers.newName(getName(),subname);
         if (container != null) {
            Element reference = container.getReference();
            if (reference == null) {
               return NameHelpers.newName("",new_name);
            } else if (reference instanceof TreeElement) {
               return ((TreeElement) reference).getAbsoluteName(new_name);
            } else {
               return NameHelpers.newName("<undefined>",new_name);
            }
         } else {
            return NameHelpers.newName("",new_name);
         }
      }

      public String toString() {
         return "TreeContextElement[" + getAbsoluteName() + " " + component + "]";
      }            
         
   }
   
}
