/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Bruno Dumant
 * 
 * --------------------------------------------------------------------------
 * $Id: PropertiesDecoder.java,v 1.1.1.1 2002/08/06 13:13:53 bmi Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonathan.libs.kernel.decoders;

import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.Element;
import org.objectweb.jonathan.apis.kernel.Component;
import org.objectweb.jonathan.apis.kernel.JonathanException;
import org.objectweb.jonathan.apis.kernel.decoders.Properties2Configuration;

import org.objectweb.jonathan.libs.kernel.JComponent;

import java.util.Properties;
import java.util.Enumeration;

/**
 * Common properties to configuration implementation.
 * May also create {@link JAtom} components.
 */
public class PropertiesDecoder implements Properties2Configuration {
   
   public PropertiesDecoder() {}

   /**
    * Imports properties in the given context, save those whose keys
    * begin with an excepted prefix. A property whose name ends with
    * <code>.atom</code> is not translated into a JProperty, but into
    * a JAtom whose element name is the remaining part of the property
    * key.
    * @param _configuration where the imported properties are added
    * @param _props the properties to import
    * @param _except the rejected property prefixes
    * @param _force forces element overwriting
    */
   public void importProperties(Context _configuration, Properties _props, String[] _except,boolean _force) {
      if (_props==null) return;
      String $currentName = null;
      String $currentValue = null;
      Element $alreadyExistingElement = null;
      for (Enumeration e = _props.propertyNames() ; e.hasMoreElements() ;) {
         $currentName = (String) e.nextElement();                
         if (reject($currentName,_except)) continue;
         $currentValue = _props.getProperty($currentName);
         if ($currentName.endsWith(".atom")) {
            try {
               String $real_name = $currentName.substring($currentName.length() - 5);
               if (!_force) {
                  Element $previous = _configuration.getElement($real_name,'.');
                  if ($previous != null) {
                     _configuration.addElement($currentName,new JAtom($currentValue),'.');
                  }
               } else {
                  _configuration.addElement($currentName,new JAtom($currentValue),'.');
               }
            } catch (JonathanException f) {
               System.err.println("Warning: Exception importing atom " + 
                                  $currentName + ".");
               f.printStackTrace();
            }
         } else {
            importProperty(_configuration,$currentName,$currentValue,_force);
         }
      }
   }

   /**
    * Imports a property in the given context.
    * @param _configuration where the imported property is added
    * @param _name the property name
    * @param _value the property value
    * @param _force forces element overwriting
    */
   public void importProperty(Context _context, String _name, String _value, boolean _force) {
      try{
         Element $previous = null;
         if (_value.equalsIgnoreCase("true")) {
            addProperty(_name,new JComponent(Boolean.class,Boolean.TRUE),_context,_force);
         } else if (_value.equalsIgnoreCase("false")) {
            addProperty(_name,new JComponent(Boolean.class,Boolean.FALSE),_context,_force);
         } else {
            try {
               // is it an integer ?
               Integer $intValue = Integer.valueOf(_value);
               addProperty(_name,new JComponent(int.class,$intValue.intValue()),_context,_force);
            } catch (NumberFormatException ignored){
               try {
                  // is it a float ?
                  Float $floatValue = Float.valueOf(_value);
                  addProperty(_name,new JComponent(Float.class,$floatValue),
                              _context,_force);
               } catch (NumberFormatException ignored2){
                  // not an integer, not a float.
                  addProperty(_name,new JComponent(String.class,_value),_context,_force);
               }
            }
         }
      } catch (JonathanException e){
         System.err.println("Warning: Exception importing property " + 
                            _name + ".");
         e.printStackTrace();
      }
   }   

   /**
    * Adds a component in the given context. Creates a new element.
    * @param _name the component's element name
    * @param _component the property component to add
    * @param _context where the property will be added
    * @param _force forces element overwriting
    */
   static void addProperty(String _name,Component _component, Context _context, boolean _force) 
      throws JonathanException {
      Element $previous = _context.addElement(_name,_component,'.');
      if (!_force && $previous != null) {
         _context.addElement(_name,$previous.getComponent(),'.');
      }
   }

   /** 
    * tests if the name has a prefix in except.
    * @param name
    * @param except
    * @return 
    */
   static boolean reject(String name, String[] except) {
      if (except == null || except.length == 0) {
         return false;
      }
      int len = except.length;
      int i = 0;
      while (i < len && ! name.startsWith(except[i])) {
         i++;
      }
      return i != len;
   }
}
