/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Bruno Dumant, Priscille Durville
 *
 * -------------------------------------------------------------------
 * $Id: XMLIncludeDecoder.java,v 1.1.1.1 2002/08/06 13:13:55 bmi Exp $
 * -------------------------------------------------------------------
 * 
 */

package org.objectweb.jonathan.libs.kernel.decoders;


import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.ContextFactory;
import org.objectweb.jonathan.apis.kernel.Component;
import org.objectweb.jonathan.apis.kernel.Element;
import org.objectweb.jonathan.apis.kernel.JonathanException;
import org.objectweb.jonathan.apis.kernel.ClassPathFinder;

import org.objectweb.jonathan.libs.kernel.JComponent;

import nanoxml.XMLElement;
import nanoxml.XMLParseException;

import java.io.InputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import java.util.Enumeration;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

/**
 * Include specific XML or properties to runtime configuration decoder.
 * Used by the {@link OpenXMLDecoder} as one of the built-in
 * XML to runtime configuration  decoders provided it provides
 * to its decoding dispatcher.
 */
public class XMLIncludeDecoder implements XMLElementDecoder {
   
   public XMLIncludeDecoder() {}
   
   /**
    * Converts a XML <code>INCLUDE</code> element to its
    * corresponding included runtime configuration
    * @param _xml_element <code>INCLUDE</code> tag to be decoded
    * @param _element_name the include's element name
    * @param _configuration where the decoded configuration will
    * be included
    * @param return the corresponding decoded runtime included
    * configuration
    */
   public Component decode(XMLElement _xml_element, String _element_name, Context _configuration,
                           XMLElementDecoderDispatcher _dispatcher) 
      throws JonathanException {
      
      String $current_tag = _xml_element.getName();
      
      if ($current_tag.equals("INCLUDE")) {
         String $file_name = _xml_element.getStringAttribute("FILE");
         if ($file_name != null) {
            if ($file_name.endsWith(".prop") ||
                $file_name.endsWith(".properties")){//include properties file
               return importProperties(_configuration,_element_name,$file_name,_dispatcher.getFinder(),
                                       _dispatcher.getContextFactory());
            } else {//include xml file
               return _dispatcher.decode(readXMLElementFromResource($file_name,_dispatcher.getFinder()),
                                         _element_name,_configuration);
            }
         } else {
            return _dispatcher.error("No file specified for include.");
         } 
      } else {
         return _dispatcher.error("XMLIncludeDecoder error: Wrong tag " + $current_tag);
      }
   }

   XMLElement readXMLElementFromResource(String _resource_name,ClassPathFinder _finder) 
      throws JonathanException {
      try {
         InputStream $is = _finder.findResource(_resource_name);
         if ($is == null) {
            throw new JonathanException("Resource " + _resource_name + " not found.");
         }
         XMLElement $result = new XMLElement();
         $result.parseFromReader(new BufferedReader(new InputStreamReader($is)));
         return $result;
      } catch (IOException e) {
         throw new JonathanException(e);
      } catch (XMLParseException e) {
         throw new JonathanException(e);
      }
   }

   private Component importProperties(Context _configuration,
                                      String _element_name,
                                      String _resource_name,
                                      ClassPathFinder _finder,
                                      ContextFactory _context_factory) 
      throws JonathanException {
      try {
         InputStream $is = _finder.findResource(_resource_name);
         if ($is == null) {
            throw new JonathanException("Resource " + _resource_name + " not found.");
         }
         Properties $properties = new Properties();
         $properties.load($is);

         Context $new_configuration = null;
         if (_element_name != null) {
            $new_configuration = _configuration.addOrGetContext(_element_name,'/');
         } else {
            $new_configuration = _context_factory.newContext();
         }

         importPropertiesInConf($properties, $new_configuration);

         return $new_configuration;
      }catch (Exception e){
         throw new JonathanException(e);
      }
   }

   private void importPropertiesInConf(Properties _prop, Context _parent) 
      throws JonathanException {
      
      boolean $ttAPlat = false;
      LinkedList $storedElements = new LinkedList();

      String $currentName = null;
      String $currentValue = null;
      for (Enumeration e = _prop.propertyNames() ; e.hasMoreElements() ;) {
         $currentName = (String) e.nextElement();
         $currentValue = _prop.getProperty($currentName);
         JComponent $component = new JComponent(String.class,$currentValue);

         if (!$ttAPlat){
            Element $previous = _parent.addElement($currentName,$component,'.');
            if ($previous != null) {//impossible mapping => do not take into account '.'
               $ttAPlat = true;
               _parent.addElement($currentName,$previous.getComponent(),'.');
               removeAndAddAllElems(_parent, $storedElements);
               _parent.addElement($currentName,$component,(char)0);
            }
            PropertyHolder $holder = new PropertyHolder();
            $holder.name = $currentName;
            $holder.component = $component;
            $storedElements.add($holder);
         } else {
            _parent.addElement($currentName,$component,(char)0);
         }
      }

      if ($ttAPlat){removeEmptyContext(_parent, null);}
   }

   private void removeAndAddAllElems(Context _parent, List _list) 
      throws JonathanException {
      for (int i=0; i<_list.size(); i++){
         PropertyHolder $current = (PropertyHolder)_list.get(i);
         String $currentName = $current.name;
         _parent.addElement($currentName, Object.class,Context.NO_VALUE,'.');
         _parent.addElement($currentName,$current.component,(char)0);
      }
   }
   
   private void removeEmptyContext(Context _configuration, Context _parent)
      throws JonathanException {
      for (Enumeration e = _configuration.getElements(); e.hasMoreElements() ;) {
         Element $current = (Element)e.nextElement();
         if ($current.getComponent() instanceof Context){
            removeEmptyContext((Context)$current.getComponent(),_configuration);
         }
      }

      if (_parent == null) {return;}
      if (_configuration.getElements().hasMoreElements()){
         //if all containing elements are Context.NO_VALUE component associated => remove it!
         for (Enumeration e = _configuration.getElements(); e.hasMoreElements() ;){
            Element $child = (Element)e.nextElement();
            if (! ($child.getComponent().getValue() == Context.NO_VALUE)){
               return;
            }
         }
      }
      Element $theElement = ((JComponent)_configuration).getReference();
      _parent.addElement($theElement.getName(),Object.class,Context.NO_VALUE,'/');
   }


   class PropertyHolder {
      String name;
      JComponent component;
   }
}
